@@ -4,15 +4,16 @@ use strict;
 use warnings;
 use v5;
 
-our $VERSION = '0.002';
+our $VERSION = '0.003';
 
 use Data::Dumper;
 use Devel::Declare;
 use Devel::Declare::Lexer::Stream;
 use Devel::Declare::Lexer::Token;
-use Devel::Declare::Lexer::Token::Comma;
+use Devel::Declare::Lexer::Token::Bareword;
 use Devel::Declare::Lexer::Token::Declarator;
 use Devel::Declare::Lexer::Token::EndOfStatement;
+use Devel::Declare::Lexer::Token::Heredoc;
 use Devel::Declare::Lexer::Token::LeftBracket;
 use Devel::Declare::Lexer::Token::Newline;
 use Devel::Declare::Lexer::Token::Operator;
@@ -112,7 +113,6 @@ sub lexer
     tie @tokens, "Devel::Declare::Lexer::Stream";
     my ($len, $tok);
     my $eoleos = 0;
-    my %lineoffsets;
     my $line = 1;
 
     # Skip the declarator
@@ -120,7 +120,11 @@ sub lexer
     push @tokens, new Devel::Declare::Lexer::Token::Declarator( value => $symbol );
     $DEBUG and say STDERR "Skipped declarator '$symbol'";
 
+    my %lineoffsets = ( 1 => $offset );
+
+    # We call this from a few places inside the loop
     my $skipspace = sub {
+        # Move past any whitespace
         $len = Devel::Declare::toke_skipspace($offset);
         if($len > 0) {
             $tok = substr($linestr, $offset, $len);
@@ -129,22 +133,20 @@ sub lexer
             $offset += $len;
 
             if($tok =~ /\n/) {
-                $DEBUG and say STDERR "Got end of line in skipspace, unusual circumstances";
-                # FIXME really?
+                # its odd that this works without handling any line numbering
+                # I think we end up here when an end of line is found after a bareword (e.g. print\n"something")
+                # It probably still needs some work on line numbering, but everything just seems to work! 
+                $DEBUG and say STDERR "Got end of line in skipspace, probable bareword preceeding EOL";
                 Devel::Declare::clear_lex_stuff;
 
+                # We've got a new line so we need to refresh our linestr
                 $linestr = Devel::Declare::get_linestr;
                 $original_linestr = $linestr;
 
-                #if($line == 1) {
-                #    $lineoffsets{1} = (length $symbol) + 1;
-                #};
-                #$line++;
-                #$lineoffsets{$line} = $offset;
-
                 $DEBUG and say STDERR "Refreshed linestr [$linestr]";
             }
         } elsif ($len < 0) {
+            # Again, its odd that we don't handle any line numbering here, and a $len of < 0 is a definite EOL
             $DEBUG and say STDERR "Got end of line in skipspace";
         } elsif ($len == 0) {
             $DEBUG and say STDERR "No whitespace skipped";
@@ -152,9 +154,37 @@ sub lexer
         return $len;
     };
 
-    # get the message
+    # Capture the tokens
     $DEBUG and say STDERR "Linestr length [", length $linestr, "]";
+    my $heredoc = undef;
+    my $heredoc_end_re = undef;
+    my $heredoc_end_re2 = undef;
     while($offset < length $linestr) {
+        $DEBUG and say STDERR Dumper \%lineoffsets;
+        if($heredoc && !(substr($linestr, $offset, 2) eq "\n")) {
+            my $c = substr($linestr, $offset, 1);
+            $DEBUG and say STDERR "Consuming char from heredoc: '$c'";
+            $offset += 1;
+            if($c =~ /\n/) {
+                $DEBUG and say STDERR "Newline found in heredoc (current line $line)";
+                #$line++;
+                #$lineoffsets{$line} = $offset;
+            } else {
+                $heredoc->{value} .= $c;
+            }
+            $DEBUG and say STDERR "New heredoc value: " . $heredoc->{value};
+            my $heredoc_name = $heredoc->{name};
+            if($heredoc->{value} =~ /$heredoc_end_re/) {
+                $heredoc->{value} =~ s/$heredoc_end_re2//;
+                $DEBUG and say STDERR "Consumed heredoc, name [$heredoc_name]:\n" . $heredoc->{value};
+                push @tokens, $heredoc;
+                $heredoc = undef;
+                $heredoc_end_re = undef;
+                $heredoc_end_re2 = undef;
+            }
+            next;
+        }
+
         $DEBUG and say STDERR "Offset[$offset], Remaining[", substr($linestr, $offset), "]";
 
         if(substr($linestr, $offset, 1) eq ';') {
@@ -166,14 +196,25 @@ sub lexer
         }
 
         if(substr($linestr, $offset, 2) eq "\n") {
-            $DEBUG and say STDERR "Got end of line in loop (current line $line)";
-            push @tokens, new Devel::Declare::Lexer::Token::Newline;
-            $offset += 1;
+            if($heredoc) {
+                $DEBUG and say STDERR "Got end of line in heredoc";
+                $heredoc->{value} .= "\n";
+            }
 
+            if(!$heredoc) {
+                $DEBUG and say STDERR "Got end of line in loop (current line $line)";
+                push @tokens, new Devel::Declare::Lexer::Token::Newline;
+                $offset += 1;
+            }
+
+            # this lets us capture a newline directly after a semicolon
+            # and immediately exit the loop - otherwise we might start
+            # consuming code that doesn't belong to us
             last if $eoleos;
             $eoleos = 0;
 
-            # we're actually consuming a new line now
+            # If we're here, it's just a new line inside the statement that 
+            # we do want to consume
 
             # We don't use skipspace here - it does too much!
             #&$skipspace;
@@ -187,21 +228,28 @@ sub lexer
 
             Devel::Declare::clear_lex_stuff;
 
+            # Got a new line, so we need to refresh linestr
             $linestr = Devel::Declare::get_linestr;
+            # It's not the next line, its everything upto and including the next line
+            # so really our original_linestr is wrong!
             $original_linestr = $linestr;
 
+            # Record some offsets for later - we start on line 1 and the first $line++ is 2
+            # so we make a special case for recording line 1's offset
             if($line == 1) {
                 $lineoffsets{1} = (length $symbol) + 1;
             };
             $line++;
-            $lineoffsets{$line} = $offset;
+            $lineoffsets{$line} = $heredoc ? $offset + 1 : $offset;
 
-            $DEBUG and say STDERR "Refreshed linestr [$linestr]";
+            $DEBUG and say STDERR "Refreshed linestr [$linestr], added lineoffset for line $line, offset $offset";
             next;
         }
 
+        # FIXME Does this ever happen?
         last if &$skipspace < 0;
 
+        # Check if its a opening bracket
         if(substr($linestr, $offset, 1) =~ /(\{|\[|\()/) {
             my $b = substr($linestr, $offset, 1);
             push @tokens, new Devel::Declare::Lexer::Token::LeftBracket( value => $b );
@@ -209,6 +257,7 @@ sub lexer
             $offset += 1;
             next;
         }
+        # Check if its a closing bracket
         if(substr($linestr, $offset, 1) =~ /(\}|\]|\))/) {
             my $b = substr($linestr, $offset, 1);
             push @tokens, new Devel::Declare::Lexer::Token::RightBracket( value => $b );
@@ -216,7 +265,7 @@ sub lexer
             $offset += 1;
             next;
         }
-
+        # Check for a reference
         if(substr($linestr, $offset, 1) =~ /\\/) {
             $tok = substr($linestr, $offset, 1);
             $DEBUG and say STDERR "Got reference operator '$tok'";
@@ -224,7 +273,7 @@ sub lexer
             $offset += 1;
             next;
         }
-
+        # Check for variable
         if(substr($linestr, $offset, 1) =~ /(\$|\%|\@|\*)/) {
             # get the sign
             # TODO the variable name is captured later - it should probably be done here
@@ -234,30 +283,20 @@ sub lexer
             $offset += 1;
             next;
         }
-
-        if(substr($linestr, $offset, 1) =~ /[!\+\-\*\/\.><=]/) {
-            $tok = substr($linestr, $offset, 1);
-            $DEBUG and say STDERR "Got operator '$tok'";
-            push @tokens, new Devel::Declare::Lexer::Token::Operator( value => $tok );
-            $offset += 1;
-            next;
-        }
-
-        if(substr($linestr, $offset, 1) eq ',') {
-            $DEBUG and say STDERR "Got a comma";
-            push @tokens, new Devel::Declare::Lexer::Token::Comma;
-            $offset += 1;
-            next;
-        }
-
+        # Check for string
         if(substr($linestr, $offset, 1) =~ /^(q|\"|\')/) {
             # FIXME need to determine string type properly
             my $strstype = substr($linestr, $offset, 1);
             my $stretype = $strstype;
             if($strstype =~ /q/) {
-                $offset += 1;
-                $strstype .= substr($linestr, $offset, 1);
-                $stretype = substr($strstype, 1);
+                if(substr($linestr, $offset, 2) =~ /qq/) {
+                    $strstype = substr($linestr, $offset, 3);
+                    $offset += 2;
+                } else {
+                    $strstype = substr($linestr, $offset, 2);
+                    $offset += 1;
+                }
+                $stretype = substr($linestr, $offset, 1);
                 $stretype =~ tr/\(/)/;
                 $len = Devel::Declare::toke_scan_str($offset);
             } else {
@@ -276,12 +315,44 @@ sub lexer
 
             next;
         }
+        # Check for heredoc
+        if(substr($linestr, $offset)=~ /^(<<\s*([\w\d]+)\s*\n)/) {
+            # Heredocs are weird - we'll just remember we're in a heredoc until we get the end token
+            $DEBUG and say STDERR "Got a heredoc with name '$2'";
+            $heredoc = new Devel::Declare::Lexer::Token::Heredoc( name => $2, value => '' );
+            $heredoc_end_re = qr/\n$2\n$/;
+            $heredoc_end_re2 = qr/$2\n$/;
+            $DEBUG and say STDERR "Created regex $heredoc_end_re and $heredoc_end_re2";
 
+            # get a new linestr - we might have captured multiple lines
+            $offset += 2 + (length $1);
+    
+            $len = Devel::Declare::toke_skipspace($offset);
+            $linestr = Devel::Declare::get_linestr;
+            $offset += $len;
+            $DEBUG and say STDERR "Skipped $len whitespace at start of heredoc, got new linestr[$linestr]";
+
+            $line++;
+            $lineoffsets{$line} = $offset;
+
+            # If we do have multiple lines, we'll fix line numbering at the end
+
+            next;
+        }
+        # Check for operator after strings (so heredocs <<NAME work)
+        if(substr($linestr, $offset, 1) =~ /[!\+\-\*\/\.><=,]/) {
+            $tok = substr($linestr, $offset, 1);
+            $DEBUG and say STDERR "Got operator '$tok'";
+            push @tokens, new Devel::Declare::Lexer::Token::Operator( value => $tok );
+            $offset += 1;
+            next;
+        }
+        # Check for bareword
         $len = Devel::Declare::toke_scan_word($offset, 1);
         if($len) {
             $tok = substr($linestr, $offset, $len);
-            $DEBUG and say STDERR "Got token '$tok'";
-            push @tokens, new Devel::Declare::Lexer::Token( value => $tok );
+            $DEBUG and say STDERR "Got bareword '$tok'";
+            push @tokens, new Devel::Declare::Lexer::Token::Bareword( value => $tok );
             $offset += $len;
             next;
         }
@@ -433,12 +504,14 @@ taking care of multiline strings or code blocks - it's all done for you.
 
 =head1 SEE ALSO
 
-For more information about how L<Devel::Declare::Lexer>works , read the 
+Some examples can be found in the source download.
+
+For more information about how L<Devel::Declare::Lexer> works, read the 
 documentation for L<Devel::Declare>.
 
 =head1 AUTHORS
 
-Ian Kent - E<lt>email@iankent.co.uk<gt> - original author
+Ian Kent - E<iankent@cpan.org> - original author
 
 http://www.iankent.co.uk/
 
